@@ -7,23 +7,7 @@
  ******************************************************************************
  * Copyright (c) 2019, TileDB, Inc
  *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice shall be included
- * in all copies or substantial portions of the Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
- * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
+ * SPDX-License-Identifier: MIT
  ****************************************************************************/
 
 #ifndef TILEDB_HEADERS_H
@@ -39,46 +23,7 @@
 #include "gdal_pam.h"
 #include "ogrsf_frmts.h"
 
-#ifdef HAVE_GCC_SYSTEM_HEADER
-#pragma GCC system_header
-#endif
-
-#ifdef _MSC_VER
-#pragma warning(push)
-#pragma warning(disable : 4996) /* XXXX was deprecated */
-#endif
-
-#include "tiledb/tiledb"
-#include "tiledb/tiledb_experimental"
-
-#ifdef _MSC_VER
-#pragma warning(pop)
-#endif
-
-#if TILEDB_VERSION_MAJOR > 2 ||                                                \
-    (TILEDB_VERSION_MAJOR == 2 && TILEDB_VERSION_MINOR >= 9)
-#define HAS_TILEDB_GROUP
-#endif
-
-#if TILEDB_VERSION_MAJOR > 2 ||                                                \
-    (TILEDB_VERSION_MAJOR == 2 && TILEDB_VERSION_MINOR >= 10)
-#define HAS_TILEDB_BOOL
-#endif
-
-#if TILEDB_VERSION_MAJOR > 2 ||                                                \
-    (TILEDB_VERSION_MAJOR == 2 && TILEDB_VERSION_MINOR >= 11)
-#define HAS_TILEDB_WORKING_OR_FILTER
-#endif
-
-#if TILEDB_VERSION_MAJOR > 2 ||                                                \
-    (TILEDB_VERSION_MAJOR == 2 && TILEDB_VERSION_MINOR >= 14)
-#define HAS_TILEDB_WORKING_UTF8_STRING_FILTER
-#endif
-
-#if TILEDB_VERSION_MAJOR > 2 ||                                                \
-    (TILEDB_VERSION_MAJOR == 2 && TILEDB_VERSION_MINOR >= 15)
-#define HAS_TILEDB_DIMENSION_LABEL
-#endif
+#include "include_tiledb.h"
 
 #if TILEDB_VERSION_MAJOR > 2 ||                                                \
     (TILEDB_VERSION_MAJOR == 2 && TILEDB_VERSION_MINOR >= 17)
@@ -89,10 +34,12 @@ struct gdal_tiledb_vector_of_bool
     bool *m_v = nullptr;
 
     gdal_tiledb_vector_of_bool() = default;
+
     ~gdal_tiledb_vector_of_bool()
     {
         std::free(m_v);
     }
+
     gdal_tiledb_vector_of_bool(gdal_tiledb_vector_of_bool &&other)
         : m_size(other.m_size), m_capacity(other.m_capacity),
           m_v(std::move(other.m_v))
@@ -101,6 +48,7 @@ struct gdal_tiledb_vector_of_bool
         other.m_capacity = 0;
         other.m_v = nullptr;
     }
+
     gdal_tiledb_vector_of_bool(const gdal_tiledb_vector_of_bool &) = delete;
     gdal_tiledb_vector_of_bool &
     operator=(const gdal_tiledb_vector_of_bool &) = delete;
@@ -111,22 +59,27 @@ struct gdal_tiledb_vector_of_bool
     {
         return m_size;
     }
+
     const bool *data() const
     {
         return m_v;
     }
+
     bool *data()
     {
         return m_v;
     }
+
     bool &operator[](size_t idx)
     {
         return m_v[idx];
     }
+
     bool operator[](size_t idx) const
     {
         return m_v[idx];
     }
+
     void resize(size_t new_size)
     {
         if (new_size > m_capacity)
@@ -146,35 +99,30 @@ struct gdal_tiledb_vector_of_bool
             memset(m_v + m_size, 0, (new_size - m_size) * sizeof(bool));
         m_size = new_size;
     }
+
     void clear()
     {
         resize(0);
     }
+
     size_t capacity() const
     {
         return m_capacity;
     }
+
     void push_back(uint8_t v)
     {
         resize(size() + 1);
         m_v[size() - 1] = static_cast<bool>(v);
     }
 };
+
 #define VECTOR_OF_BOOL gdal_tiledb_vector_of_bool
 #define VECTOR_OF_BOOL_IS_NOT_UINT8_T
 #else
 #define VECTOR_OF_BOOL std::vector<uint8_t>
 #endif
 
-#ifdef HAS_TILEDB_DIMENSION_LABEL
-#define HAS_TILEDB_MULTIDIM
-#endif
-
-#if TILEDB_VERSION_MAJOR > 2 ||                                                \
-    (TILEDB_VERSION_MAJOR == 2 && TILEDB_VERSION_MINOR >= 21)
-#define HAS_TILEDB_GEOM_WKB_WKT
-#endif
-
 typedef enum
 {
     BAND = 0,
@@ -190,6 +138,11 @@ constexpr const char *TILEDB_VALUES = "TDB_VALUES";
 
 constexpr const char *GDAL_ATTRIBUTE_NAME = "_gdal";
 
+constexpr const char *DATASET_TYPE_ATTRIBUTE_NAME = "dataset_type";
+// Potential values for dataset_type metadata:
+constexpr const char *RASTER_DATASET_TYPE = "raster";
+constexpr const char *GEOMETRY_DATASET_TYPE = "geometry";
+
 /************************************************************************/
 /* ==================================================================== */
 /*                               TileRasterBand                         */
@@ -226,13 +179,12 @@ class TileDBDataset : public GDALPamDataset
                                    char **papszOptions,
                                    GDALProgressFunc pfnProgress,
                                    void *pProgressData);
-#ifdef HAS_TILEDB_MULTIDIM
+
     static GDALDataset *OpenMultiDimensional(GDALOpenInfo *);
     static GDALDataset *
     CreateMultiDimensional(const char *pszFilename,
                            CSLConstList papszRootGroupOptions,
                            CSLConstList papszOptions);
-#endif
 };
 
 /************************************************************************/
@@ -246,11 +198,14 @@ class TileDBRasterDataset final : public TileDBDataset
     friend class TileDBRasterBand;
 
   protected:
+    std::string m_osConfigFilename{};
     std::unique_ptr<tiledb::Context> m_roCtx;
     std::unique_ptr<tiledb::Array> m_array;
     std::unique_ptr<tiledb::Array> m_roArray;
     std::unique_ptr<tiledb::ArraySchema> m_schema;
     std::unique_ptr<tiledb::FilterList> m_filterList;
+    bool m_bDatasetInGroup = false;
+    std::string m_osArrayURI{};
     CPLString osMetaDoc;
     TILEDB_INTERLEAVE_MODE eIndexMode = BAND;
     int nBitsPerSample = 8;
@@ -260,34 +215,67 @@ class TileDBRasterDataset final : public TileDBDataset
     int nBlocksX = 0;
     int nBlocksY = 0;
     uint64_t nBandStart = 1;
-    bool bHasSubDatasets = false;
-    int nSubDataCount = 0;
-    char **papszSubDatasets = nullptr;
-    CPLStringList m_osSubdatasetMD{};
-    CPLXMLNode *psSubDatasetsTree = nullptr;
-    char **papszAttributes = nullptr;
-    std::list<std::unique_ptr<GDALDataset>> lpoAttributeDS = {};
+    bool m_bHasSubDatasets = false;
+    CPLStringList m_aosSubdatasetMD{};
+    CPLXMLTreeCloser m_poSubDatasetsTree{nullptr};
+    std::list<std::unique_ptr<GDALDataset>> m_lpoAttributeDS = {};
     uint64_t nTimestamp = 0;
-
     bool bStats = FALSE;
+    bool m_bDeferredCreateHasRun = false;
+    bool m_bDeferredCreateHasBeenSuccessful = false;
+
+    //! Number of overviews declared in _gdal metadata. In theory, it should
+    // match m_apoOverviewDS.size(), but do not strongly rely on that.
+    int m_nOverviewCountFromMetadata = 0;
+
+    //! Overview datasets
+    std::vector<std::unique_ptr<GDALDataset>> m_apoOverviewDS{};
+
+    //! Overview datasets that have been removed per IBuildOverviews(nOverview==0)
+    std::vector<std::unique_ptr<GDALDataset>> m_apoOverviewDSRemoved{};
+
+    //! Whether LoadOverviews() has already been called.
+    bool m_bLoadOverviewsDone = false;
+
     CPLErr IRasterIO(GDALRWFlag, int, int, int, int, void *, int, int,
-                     GDALDataType, int, int *, GSpacing, GSpacing, GSpacing,
-                     GDALRasterIOExtraArg *psExtraArg) override;
+                     GDALDataType, int, BANDMAP_TYPE, GSpacing, GSpacing,
+                     GSpacing, GDALRasterIOExtraArg *psExtraArg) override;
     CPLErr CreateAttribute(GDALDataType eType, const CPLString &osAttrName,
-                           const int nSubRasterCount = 1);
+                           const int nSubRasterCount, bool bHasFillValue,
+                           double dfFillValue);
 
     CPLErr AddDimensions(tiledb::Domain &domain, const char *pszAttrName,
                          tiledb::Dimension &y, tiledb::Dimension &x,
                          tiledb::Dimension *poBands = nullptr);
 
+    void CreateArray();
+    bool DeferredCreate(bool bCreateArray);
+
+    tiledb::Array &GetArray(bool bForWrite, tiledb::Context *&ctx);
+
+    static GDALDataset *OpenInternal(GDALOpenInfo *,
+                                     tiledb::Object::Type objectType);
+
+    //! Load TileDB overviews from TileDB arrays
+    void LoadOverviews();
+
   public:
     ~TileDBRasterDataset();
-    CPLErr TryLoadCachedXML(char **papszSiblingFiles = nullptr,
+    CPLErr TryLoadCachedXML(CSLConstList papszSiblingFiles = nullptr,
                             bool bReload = true);
-    CPLErr TryLoadXML(char **papszSiblingFiles = nullptr) override;
+    CPLErr TryLoadXML(CSLConstList papszSiblingFiles = nullptr) override;
     CPLErr TrySaveXML() override;
     char **GetMetadata(const char *pszDomain) override;
+    CPLErr Close() override;
+    int CloseDependentDatasets() override;
     virtual CPLErr FlushCache(bool bAtClosing) override;
+
+    CPLErr IBuildOverviews(const char *pszResampling, int nOverviews,
+                           const int *panOverviewList, int nListBands,
+                           const int *panBandList, GDALProgressFunc pfnProgress,
+                           void *pProgressData,
+                           CSLConstList papszOptions) override;
+
     static CPLErr CopySubDatasets(GDALDataset *poSrcDS,
                                   TileDBRasterDataset *poDstDS,
                                   GDALProgressFunc pfnProgress,
@@ -295,13 +283,13 @@ class TileDBRasterDataset final : public TileDBDataset
     static TileDBRasterDataset *CreateLL(const char *pszFilename, int nXSize,
                                          int nYSize, int nBands,
                                          GDALDataType eType,
-                                         char **papszOptions);
-    static void SetBlockSize(GDALRasterBand *poBand, char **&papszOptions);
+                                         CSLConstList papszOptions);
+    static void SetBlockSize(GDALRasterBand *poBand, CPLStringList &aosOptions);
 
-    static GDALDataset *Open(GDALOpenInfo *);
-    static GDALDataset *Create(const char *pszFilename, int nXSize, int nYSize,
-                               int nBands, GDALDataType eType,
-                               char **papszOptions);
+    static GDALDataset *Open(GDALOpenInfo *, tiledb::Object::Type objectType);
+    static TileDBRasterDataset *Create(const char *pszFilename, int nXSize,
+                                       int nYSize, int nBands,
+                                       GDALDataType eType, char **papszOptions);
     static GDALDataset *CreateCopy(const char *pszFilename,
                                    GDALDataset *poSrcDS, int bStrict,
                                    char **papszOptions,
@@ -334,6 +322,7 @@ class OGRTileDBLayer final : public OGRLayer,
 
   private:
     friend OGRTileDBDataset;
+    GDALDataset *m_poDS = nullptr;
     std::string m_osGroupName{};
     std::string m_osFilename{};
     uint64_t m_nTimestamp = 0;
@@ -368,11 +357,7 @@ class OGRTileDBLayer final : public OGRLayer,
     uint64_t m_nRowCountInResultSet = 0;
     int m_nUseOptimizedAttributeFilter = -1;  // uninitialized
 
-#ifdef HAS_TILEDB_WORKING_UTF8_STRING_FILTER
     tiledb_datatype_t m_eTileDBStringType = TILEDB_STRING_UTF8;
-#else
-    tiledb_datatype_t m_eTileDBStringType = TILEDB_STRING_ASCII;
-#endif
 
     std::string m_osXDim = "_X";
     std::string m_osYDim = "_Y";
@@ -484,18 +469,19 @@ class OGRTileDBLayer final : public OGRLayer,
                           struct ArrowArray *out_array) override;
 
   public:
-    OGRTileDBLayer(const char *pszFilename, const char *pszLayerName,
-                   const OGRwkbGeometryType eGType,
+    OGRTileDBLayer(GDALDataset *poDS, const char *pszFilename,
+                   const char *pszLayerName, const OGRwkbGeometryType eGType,
                    const OGRSpatialReference *poSRS);
     ~OGRTileDBLayer();
     void ResetReading() override;
-    DEFINE_GET_NEXT_FEATURE_THROUGH_RAW(OGRTileDBLayer);
+    DEFINE_GET_NEXT_FEATURE_THROUGH_RAW(OGRTileDBLayer)
     OGRFeature *GetFeature(GIntBig nFID) override;
     OGRErr ICreateFeature(OGRFeature *poFeature) override;
-    OGRErr CreateField(OGRFieldDefn *poField, int bApproxOK) override;
+    OGRErr CreateField(const OGRFieldDefn *poField, int bApproxOK) override;
     int TestCapability(const char *) override;
     GIntBig GetFeatureCount(int bForce) override;
     OGRErr GetExtent(OGREnvelope *psExtent, int bForce = TRUE) override;
+
     OGRErr GetExtent(int iGeomField, OGREnvelope *psExtent, int bForce) override
     {
         return OGRLayer::GetExtent(iGeomField, psExtent, bForce);
@@ -505,6 +491,7 @@ class OGRTileDBLayer final : public OGRLayer,
     {
         return m_osFIDColumn.c_str();
     }
+
     OGRFeatureDefn *GetLayerDefn() override
     {
         return m_poFeatureDefn;
@@ -514,6 +501,11 @@ class OGRTileDBLayer final : public OGRLayer,
 
     const char *GetMetadataItem(const char *pszName,
                                 const char *pszDomain) override;
+
+    GDALDataset *GetDataset() override
+    {
+        return m_poDS;
+    }
 };
 
 /************************************************************************/
@@ -532,20 +524,24 @@ class OGRTileDBDataset final : public TileDBDataset
     OGRLayer *ExecuteSQL(const char *pszSQLCommand,
                          OGRGeometry *poSpatialFilter,
                          const char *pszDialect) override;
+
     int GetLayerCount() override
     {
         return static_cast<int>(m_apoLayers.size());
     }
+
     OGRLayer *GetLayer(int nIdx) override
     {
         return nIdx >= 0 && nIdx < GetLayerCount() ? m_apoLayers[nIdx].get()
                                                    : nullptr;
     }
+
     int TestCapability(const char *) override;
+
     OGRLayer *ICreateLayer(const char *pszName,
-                           const OGRSpatialReference *poSpatialRef = nullptr,
-                           OGRwkbGeometryType eGType = wkbUnknown,
-                           char **papszOptions = nullptr) override;
+                           const OGRGeomFieldDefn *poGeomFieldDefn,
+                           CSLConstList papszOptions) override;
+
     static GDALDataset *Open(GDALOpenInfo *, tiledb::Object::Type objectType);
     static GDALDataset *Create(const char *pszFilename,
                                CSLConstList papszOptions);
