@@ -9,23 +9,7 @@
  ******************************************************************************
  * Copyright (c) 2016, Even Rouault, <even dot rouault at spatialys dot com>
  *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice shall be included
- * in all copies or substantial portions of the Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
- * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
+ * SPDX-License-Identifier: MIT
  ****************************************************************************/
 
 #include "ogr_gmlas.h"
@@ -111,7 +95,7 @@ CPLString OGRGMLASTruncateIdentifier(const CPLString &osName,
     {
         if (nLastTokenSize > nExtra)
         {
-            aosTokens[aosTokens.size() - 1].resize(nLastTokenSize - nExtra);
+            aosTokens.back().resize(nLastTokenSize - nExtra);
             nExtra = 0;
         }
     }
@@ -127,7 +111,7 @@ CPLString OGRGMLASTruncateIdentifier(const CPLString &osName,
             // than previous ones.
             if (nLastTokenSize > 2 * iter->first)
             {
-                aosTokens[aosTokens.size() - 1].resize(nLastTokenSize - 1);
+                aosTokens.back().resize(nLastTokenSize - 1);
                 nLastTokenSize--;
                 bHasDoneSomething = true;
                 nExtra--;
